/* 
   PXKMenu.m

   NSMenu for GNUstep GUI X/DPS Backend

   Copyright (C) 1996 Free Software Foundation, Inc.

   Author:  Scott Christley <scottc@net-community.com>
   Date: March 1996
   
   This file is part of the GNUstep GUI X/DPS Backend.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   If you are interested in a warranty or support for this source code,
   contact Scott Christley <scottc@net-community.com> for more information.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 

#include <gnustep/xdps/PXKMenu.h>
#include <gnustep/xdps/PXKDPSContext.h>
#include <gnustep/xdps/PXKScreen.h>
#include <gnustep/xdps/PXKColor.h>
#include <AppKit/NSApplication.h>
#include <gnustep/xdps/PXKWindow.h>

typedef struct _PXKMenu_struct
{
  Window xWindow;
  GC context;
  BOOL is_exposed;
  BOOL needs_calc;
  NSPoint root_origin;
  BOOL valid_origin;
} PXKMenu_struct;

#define XWINDOW (((PXKMenu_struct *)be_menu_reserved)->xWindow)
#define XGC (((PXKMenu_struct *)be_menu_reserved)->context)
#define XEXPOSE (((PXKMenu_struct *)be_menu_reserved)->is_exposed)
#define XCALC (((PXKMenu_struct *)be_menu_reserved)->needs_calc)
#define PXKROOT (((PXKMenu_struct *)be_menu_reserved)->root_origin)
#define PXKVALID (((PXKMenu_struct *)be_menu_reserved)->valid_origin)

@implementation PXKMenu

//
// Initializing a New NSMenu 
//
- initWithTitle:(NSString *)aTitle
{
  // Allocate the back-end structure
  be_menu_reserved = malloc(sizeof(PXKMenu_struct));
  XWINDOW = 0;
  XCALC = YES;
  XEXPOSE = NO;
  PXKROOT = NSZeroPoint;
  PXKVALID = NO;

  // Use 24 point as default for 1/4 inch high
  // plus offset of 8
  frame.size.width = 200;
  frame.size.height = 33;

  [super initWithTitle: aTitle];

  needs_display = YES;

  return self;
}

- (void)dealloc;
{
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
					xDisplay];

  // Destroy the X Window
  if (XWINDOW)
    {
      XDestroyWindow(xDisplay, XWINDOW);
      XWINDOW = 0;
    }

  // Release the back-end structure
  if (be_menu_reserved)
    free(be_menu_reserved);

  [super dealloc];
}

- (void)closeSubmenu:(Display *)xDisplay
{
  id e;
  NSMenuCell *mc;

  e = [menu_items objectEnumerator];
  mc = [e nextObject];
  while (mc)
    {
      if ([mc hasSubmenu])
	[(PXKMenu *)[mc submenu] closeSubmenu: xDisplay];
      mc = [e nextObject];
    }

  if (XWINDOW)
    XUnmapWindow(xDisplay, XWINDOW);
}

- (void)close;
{
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
					xDisplay];
  [self closeSubmenu: xDisplay];
  needs_display = YES;
}

- (void)display;
{
  NSDebugLog(@"PXKMenu -display\n");

  // If we are a submenu
  if (super_menu)
    {
      // Calculate the size of the menu if need be
      if (XCALC)
	{
	  [self sizeToFit];

	  // Add the last row to the matrix
	  [menu_matrix addColumnWithCells: menu_items];

	  XCALC = NO;
	}

      // Create the associated X Window
      [self createXWindow: frame];
    }
  else
    // The top level menu gets drawn elsewhere
    return;

  if (!XEXPOSE)
    return;

  [content_view display];

  needs_display = NO;
}

- (void)displayIfNeeded
{
  NSDebugLog(@"PXKMenu displayIfNeeded\n");
  if (needs_display)
    [self display];
}

//
// Managing NSMenu Windows 
//
- (NSPoint)locationForSubmenu:(NSMenu *)aSubmenu
{
  id e;
  NSMenuCell *mc;
  BOOL found;
  NSRect cellFrame = NSZeroRect;
  NSPoint cp = NSZeroPoint;
  int aRow, aCol;

  NSDebugLog(@"PXKMenu -locationForSubmenu:\n");

  // bogus answer if you don't specify a submenu
  if (!aSubmenu)
    return NSZeroPoint;

  // Find the submenu
  e = [menu_items objectEnumerator];
  mc = (NSMenuCell *)[e nextObject];
  found = NO;
  while ((mc) && (!found))
    {
      if ([mc submenu] == aSubmenu)
	found = YES;
      else
	mc = (NSMenuCell *)[e nextObject];
    }

  // If we didn't find it then well?
  // +++ Are we suppose to handle this case?
  if (!found)
    {
      NSDebugLog(@"PXKMenu -locationForSubmenu: did not find submenu\n");
      return NSZeroPoint;
    }

  // Get the cell frame
  [menu_matrix getRow: &aRow column: &aCol ofCell: mc];
  NSDebugLog(@"PXKMenu row %d col %d\n", aRow, aCol);

  cellFrame = [menu_matrix cellFrameAtRow: aRow column: aCol];
  NSDebugLog(@"PXKMenu cell frame %f %f %f %f\n", cellFrame.origin.x,
	     cellFrame.origin.y, cellFrame.size.width, cellFrame.size.height);

  // If we are the top-level menu
  // Then the calculation is slightly different
  if (!super_menu)
    {
      if (!PXKVALID)
	{
	  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
						xDisplay];
	  int xScreen = [(PXKScreen *)[NSScreen mainScreen] xScreen];
	  int rx, ry;
	  float a, b;
	  Window child;

	  // Get device coordinate of origin
	  // then translate to user coordinates
	  XTranslateCoordinates(xDisplay, XWINDOW, 
				RootWindow(xDisplay, xScreen),
				0, 0, &rx, &ry, &child);
	  NSDebugLog(@"translate to root %d %d\n", rx, ry);
	  PSidtransform(rx, ry, &a, &b);
	  if (a < 0) a = -a;
	  if (b < 0) b = -b;
	  PXKROOT.x = a;
	  PXKROOT.y = b;
	  PXKVALID = YES;
	}
      // translate to bottom right corner of cell
      cp.x = PXKROOT.x + cellFrame.origin.x;
      cp.y = PXKROOT.y + cellFrame.size.height;
    }
  else
    {
      // Traverse up the menu hierarchy
      cp = [super_menu locationForSubmenu: self];
      // translate for the menu cell
      cp.x += cellFrame.size.width;
      cp.y = cp.y + (frame.size.height - cellFrame.origin.y) - 
	cellFrame.size.height;
    }
  return cp;
}

- (NSPoint)topLevelMenuOrigin:(NSMenu *)submenu
{
  if (super_menu)
    return [(PXKMenu *)super_menu topLevelMenuOrigin: self];
  else
    {
      NSPoint cp;
      if (!PXKVALID)
	{
	  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
						xDisplay];
	  int xScreen = [(PXKScreen *)[NSScreen mainScreen] xScreen];
	  int rx, ry;
	  float a, b;
	  Window child;

	  // Get device coordinate of origin 
	  // then translate to user coordinates
	  XTranslateCoordinates(xDisplay, XWINDOW, 
				RootWindow(xDisplay, xScreen),
				0, 0, &rx, &ry, &child);
	  NSDebugLog(@"translate to root %d %d\n", rx, ry);
	  PSidtransform(rx, ry, &a, &b);
	  if (a < 0) a = -a;
	  if (b < 0) b = -b;
	  PXKROOT.x = a;
	  PXKROOT.y = b;
	  PXKVALID = YES;
	}
      cp.x = PXKROOT.x;
      cp.y = PXKROOT.y;
      // flip to get bottom right corner
      cp.y += frame.size.height;
      return cp;
    }
}

- (void)sizeToFit
{
  // Presumably the top-level menu is embedded within the window
  // So its sizing is different from submenus which pop up
  if (super_menu)
    {
      float max_width;
      float tot_height;
      float width, height;
      const char *c;
      NSFont *f;
      id e;
      NSMenuCell *o, *mc;
      NSRect cframe = frame;
      BOOL found;

      // Save the graphics state
      // Because we will change the font
      //	  [content_view lockFocus];
      PSgsave();

      // If we are the submenu of the top-level menu
      // then we want to adjust our width in accordance to
      // the top-level menu cell.
      if ([super_menu supermenu] == nil)
	{
	  // Find the supermenu's cell for ourselves
	  e = [[super_menu itemArray] objectEnumerator];
	  mc = (NSMenuCell *)[e nextObject];
	  found = NO;
	  while ((mc) && (!found))
	    {
	      if ([mc submenu] == self)
		found = YES;
	      else
		mc = (NSMenuCell *)[e nextObject];
	    }

	  if (found)
	    {
	      int aRow, aCol;
	      NSRect cellFrame;

	      // Get the cell frame which is initial width of submenu
	      [[super_menu itemMatrix] getRow: &aRow column: &aCol ofCell: mc];
	      cellFrame = [[super_menu itemMatrix] cellFrameAtRow: aRow 
						   column: aCol];
	      max_width = cellFrame.size.width;
	    }
	  else
	    {
	      // This shouldn't happen!
	      NSLog(@"PXKMenu -sizeToFit, could not find supermenu's cell\n");
	      max_width = 0;
	    }
	}
      else
	// Otherwise our width is based purely on our menu cells
	max_width = 0;

      // Loop through the menu items
      e = [menu_items objectEnumerator];
      o = (NSMenuCell *)[e nextObject];
      tot_height = 0;
      while (o)
	{
	  f = [o font];
	  [f set];

	  // Add in the height
	  tot_height += [f pointSize] + 10;

	  // Calculate the width
	  c = [[o stringValue] cString];
	  PSstringwidth(c, &width, &height);

	  if ((width + 10) > max_width)
	    max_width = width + 10;

	  NSDebugLog(@"title %s width %f\n", c, width);

	  o = (NSMenuCell *)[e nextObject];
	}

      cframe.origin = NSZeroPoint;
      cframe.size.width = max_width;
      cframe.size.height = tot_height;
      frame.size = cframe.size;
      [content_view setFrame: cframe];
      //      [menu_matrix setFrame: cframe];

      // Restore the graphics state
      //	  [content_view unlockFocus];
      PSgrestore();
    }
}

- (void)setFrame:(NSRect)frameRect display:(BOOL)flag;
{
  frame = frameRect;

  if (!flag) return;
  [self display];
}

//
// Handle events
//
- (NSPoint)convertPointToMenu:(NSPoint)p
{
  NSPoint cp, tp;

  // If we are the top-level menu
  // then the point should be in our coordinate system already
  if (!super_menu)
    return p;

  // Get our origin from our super menu
  cp = [super_menu locationForSubmenu: self];
  NSDebugLog(@"PXKMenu menu origin %f %f\n", cp.x, cp.y);

  // Get the top level menu origin
  tp = [self topLevelMenuOrigin: self];
  NSDebugLog(@"PXKMenu toplevel %f %f\n", tp.x, tp.y);

  // Translate the point
  // from the top level menu origin
  cp.x -= tp.x;
  cp.y -= tp.y;
  // from the specified point
  cp.x = p.x - cp.x;
  cp.y += p.y;
  // flip
  cp.y += frame.size.height;

  return cp;
}

- (NSMenuCell *)submenuTracking:(NSEvent *)theEvent
{
  NSPoint p;
  NSMenuCell *current_sel, *newc;
  int aRow, aCol;

  // Find the cell that is currently selected
  current_sel = [menu_matrix selectedCell];

  // Convert point to our menu
  p = [theEvent locationInWindow];
  p = [self convertPointToMenu: p];
  NSDebugLog(@"PXKMenu converted sub point %f %f\n", p.x, p.y);

  // Is it within our matrix
  if ([menu_matrix getRow: &aRow column: &aCol forPoint: p])
    {
      newc = [menu_matrix cellAtRow:aRow column:aCol];
      if (newc != current_sel)
	{
	  NSDebugLog(@"New cell at %d %d\n", aRow, aCol);
	  // Unhighlight the old cell
	  if (current_sel)
	    {
	      int oldRow, oldCol;
	      [menu_matrix getRow: &oldRow column: &oldCol 
			   ofCell: current_sel];
	      [menu_matrix highlightCell: NO atRow: oldRow column: oldCol];

	      // Close the submenu window if it has one
	      if ([current_sel hasSubmenu])
		[[current_sel submenu] close];
	    }
		  
	  // Deselect the old cell and select the new
	  [menu_matrix deselectSelectedCell];
	  [menu_matrix selectCellAtRow: aRow column: aCol];
	  current_sel = newc;

	  // highlight the new cell
	  [menu_matrix highlightCell: YES atRow: aRow column: aCol];

	  if ([current_sel hasSubmenu])
	    {
	      NSPoint subOrigin = NSZeroPoint;

	      subOrigin = [self locationForSubmenu: [current_sel submenu]];

	      NSDebugLog(@"PXKMenu submenu origin %f %f\n", 
			 subOrigin.x, subOrigin.y);
	      [[current_sel submenu] setFrameOrigin: subOrigin];
	      if (![[current_sel submenu] isVisible])
		{
		  NSDebugLog(@"PXKMenu display submenu\n");
		  [[current_sel submenu] display];
		}
	    }
	}
      return current_sel;
    }
  else
    {
      // If the current selection has a submenu
      // Then let it check
      if ([current_sel hasSubmenu])
	return [(PXKMenu *)[current_sel submenu] submenuTracking: theEvent];
      else
	{
	  // Otherwise there should be no selection
	  if (current_sel)
	    {
	      [menu_matrix deselectSelectedCell];
	      [menu_matrix getRow: &aRow column: &aCol 
			   ofCell: current_sel];
	      [menu_matrix highlightCell: NO atRow: aRow column: aCol];
	    }
	}
    }
  return nil;
}

- (void)doMenuTracking:(NSEvent *)theEvent
{
  NSApplication *theApp = [NSApplication sharedApplication];
  NSPoint p;
  BOOL done;
  int aRow, aCol;
  int newRow, newCol;
  NSMenuCell *mc = nil, *newc = nil, *selected_cell = nil;
  unsigned int event_mask = NSLeftMouseDownMask | NSLeftMouseUpMask |
    NSMouseMovedMask | NSLeftMouseDraggedMask |
    NSRightMouseDraggedMask;
  NSEvent *e;
  BOOL mouseUp = NO;

  // Convert the point to our local coordinates
  p = [theEvent locationInWindow];
  NSDebugLog(@"Got a mouse down at %f %f %f\n", p.x, p.y);
  p = [self convertPointToMenu: p];
  NSDebugLog(@"Point in my coordinate %f %f %f\n", p.x, p.y);

  // Ask the matrix for the cell at point
  if ([menu_matrix getRow: &aRow column: &aCol forPoint: p])
    {
      // If we are the top-level menu
      // then grab the mouse pointer
      if (!super_menu)
	[self captureMouse: self];

      done = NO;
      mc = [menu_matrix cellAtRow:aRow column:aCol];
      if (!mc)
	{
	  // Yikes problem here
	  NSLog(@"PXKMenu no cell at %d %d\n", aRow, aCol);
	  return;
	}

      // Highlight the cell
      [menu_matrix highlightCell: YES atRow: aRow column: aCol];
      selected_cell = mc;
    }
  else
    // If the menu wasn't hit then just return
    return;

  // Start tracking
  while (!done)
    {
      // If the menu cell has a submenu
      // then display it
      if ([mc hasSubmenu])
	{
	  NSPoint subOrigin = NSZeroPoint;

	  subOrigin = [self locationForSubmenu: [mc submenu]];

	  NSDebugLog(@"PXKMenu submenu origin %f %f\n", 
		     subOrigin.x, subOrigin.y);
	  [[mc submenu] setFrameOrigin: subOrigin];
	  if (![[mc submenu] isVisible])
	    {
	      NSDebugLog(@"PXKMenu display submenu\n");
	      [[mc submenu] display];
	    }
	}

      // Get the next event
      e = [theApp nextEventMatchingMask:event_mask untilDate:nil 
		  inMode:nil dequeue:YES];

      // If mouse went up then we are done
      if ([e type] == NSLeftMouseUp)
	{
	  done = YES;
	  mouseUp = YES;
	}

      // Convert point to our menu
      p = [e locationInWindow];
      p = [self convertPointToMenu: p];
      NSDebugLog(@"PXKMenu converted point %f %f\n", p.x, p.y);

      // Is it within our matrix
      if ([menu_matrix getRow: &newRow column: &newCol forPoint: p])
	{
	  newc = [menu_matrix cellAtRow:newRow column:newCol];
	  if (newc != mc)
	    {
	      NSDebugLog(@"New cell at %d %d\n", aRow, aCol);
	      // Unhighlight the old cell
	      [menu_matrix highlightCell: NO atRow: aRow column: aCol];
		  
	      // Close the submenu window if it has one
	      if ([mc hasSubmenu])
		[[mc submenu] close];

	      // highlight the new cell
	      mc = newc;
	      aRow = newRow;
	      aCol = newCol;
	      [menu_matrix highlightCell: YES atRow: aRow column: aCol];
	    }
	}
      else
	{
	  // Is it within the submenu matrix
	  NSDebugLog(@"Not in the main menu matrix\n");
	  if ([mc hasSubmenu])
	    {
	      newc = [(PXKMenu *)[mc submenu] submenuTracking: e];
	      if (newc)
		selected_cell = newc;
	    }
	}
    }

  // Unhighlight the cell
  [menu_matrix highlightCell: NO atRow: aRow column: aCol];

  // Close the submenu window if it has one
  if ([mc hasSubmenu])
    [[mc submenu] close];

  // If we are the top-level menu
  // then release the mouse pointer
  if (!super_menu)
    [self releaseMouse: self];

  // If mouse up
  // Then have the target perform the action
  if (mouseUp)
    {
      if (selected_cell)
	[theApp sendAction:[selected_cell action] to:[selected_cell target]
		from: selected_cell];
    }		
}

- (void)sendEvent:(NSEvent *)theEvent
{
  switch ([theEvent type])
    {
      //
      // Left mouse down
      //
    case NSLeftMouseDown:
      {
	[self doMenuTracking: theEvent];
	break;
      }
    default:
      [super sendEvent: theEvent];
    }
}

@end

//
// Methods for XWindows implementation
//
@implementation PXKMenu (GNUstepXDPS)

- (void)drawMenuForWindow:(NSWindow *)w
{
  NSRect cf = NSZeroRect;
  NSRect r = [w frame];

  XWINDOW = [(PXKWindow *)w xMenuWindow];
  XGC = [(PXKWindow *)w xMenuGC];
  XEXPOSE = YES;

  if (XCALC)
    {
      float width, height;
      const char *c;
      NSFont *f;
      id e;
      NSMenuCell *o;
      NSArray *a = [NSMutableArray array];

      frame.size.width = r.size.width;
      [content_view setFrame: frame];
      //      [menu_matrix setFrame: frame];

      [content_view lockFocus];
      // Loop through the menu items
      e = [menu_items objectEnumerator];
      o = (NSMenuCell *)[e nextObject];
      cf.origin.x = 0;
      cf.origin.y = 0;
      while (o)
	{
	  f = [o font];
	  [f set];

	  // Calculate the width
	  c = [[o stringValue] cString];
	  PSstringwidth(c, &width, &height);

	  // If the width of the item extends past the frame width
	  if ((width + cf.origin.x + 8) > (frame.size.width))
	    {
	      // add the cell array to the matrix
	      // and go to the next row
	      [menu_matrix addRowWithCells: a];
	      break;
	    }
	  else
	    {
	      // Else add the cell to the row array
	      NSLog(@"title %s width %f at %f %f %f %f\n", c, 
		    width, cf.origin.x, 
		    cf.origin.y, cf.size.width, cf.size.height);

	      [a addObject: o];
	    }

	  cf.size.width = width + 8;
	  cf.origin.x = cf.origin.x + cf.size.width;
	  o = (NSMenuCell *)[e nextObject];
	}
      // Add the last row to the matrix
      [menu_matrix addRowWithCells: a];

      [content_view unlockFocus];

      XCALC = NO;
    }
  [content_view display];
  visible = YES;
}

- (void)createXWindow:(NSRect) rect
{
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
					xDisplay];

  // Creation of a window depends upon whether the menu is the top-level
  // menu or a submenu.  The top-level menu does not create its own window
  // because each PXKWindow maintains a menu bar window which PXKMenu 
  // can draw in.  Submenus, however, must create their own window.

  // If we are a sub-menu
  if ((super_menu) && (!XWINDOW))
    {
      XGCValues values;
      unsigned long valuemask = (GCForeground | GCBackground);
      int xScreen = [(PXKScreen *)[NSScreen mainScreen] xScreen];
      XColor c1;
      XSetWindowAttributes attr;
      float a, b, c, d;

      // If not a colorspace we can easily convert then
      // then just make the background gray.
      c1.red = 32000;
      c1.green = 32000;
      c1.blue = 32000;

      // Translate the rect from user to device coordinates
      PSdtransform(rect.origin.x, rect.origin.y, &a, &b);
      PSdtransform(rect.size.width, rect.size.height, &c, &d);
      if (a < 0) a = -a;
      if (b < 0) b = -b;
      if (c < 0) c = -c;
      if (d < 0) d = -d;

      /* Create the X window */
      XWINDOW = XCreateSimpleWindow(xDisplay, RootWindow(xDisplay, xScreen),
				    a, b, c, d,
				    0, BlackPixel(xDisplay, xScreen),
				    c1.pixel);

      /* Set the event mask */
      XSelectInput(xDisplay, XWINDOW,
		   ExposureMask | KeyPressMask | KeyReleaseMask |
		   ButtonPressMask | ButtonReleaseMask | 
		   ButtonMotionMask | StructureNotifyMask |
		   PointerMotionMask | EnterWindowMask |
		   LeaveWindowMask | StructureNotifyMask | FocusChangeMask |
		   PropertyChangeMask | ColormapChangeMask | KeymapStateMask |
		   VisibilityChangeMask);

      attr.override_redirect = True;
      XChangeWindowAttributes(xDisplay, XWINDOW, CWOverrideRedirect, &attr);

      // Create a GC for the menu
      values.foreground = BlackPixel(xDisplay, xScreen);
      values.background = WhitePixel(xDisplay, xScreen);
      XGC = XCreateGC(xDisplay, XWINDOW, valuemask, &values);
    }

  // If its not visible then we should map it
  if (!visible)
    {
      NSDebugLog(@"PXKMenu map window\n");
      XMapRaised(xDisplay, XWINDOW);
      visible = YES;
    }
}

- (void)setXWindow:(Window)w
{
  XWINDOW = w;
}

- (Window)xWindow
{
  return XWINDOW;
}

- (GC)xGC
{
  return XGC;
}

- (Window)xParentWindow;
{
  return XWINDOW;
}

- (Window)xMenuWindow;
{
  return XWINDOW;
}

- (GC)xMenuGC
{
  return XGC;
}

- (void)setXExposed:(BOOL)flag
{
  NSDebugLog(@"PXKMenu xexpose %d\n", flag);
  XEXPOSE = flag;
}

- (BOOL)isXExposed
{
  return XEXPOSE;
}

- (void)setFrameFromXFrame:(NSRect)rect
{
  float x, y;

  // Adjust the frame to user coordinates
  PSidtransform(rect.size.width, rect.size.height, &x, &y);
  if (x < 0) x = -x;
  if (y < 0) y = -y;
  NSDebugLog(@"transform %f %f %f %f\n", rect.size.width, 
	     rect.size.width, x, y);
  frame.size.width = x;
  frame.size.height = y;

  // Inform our content view
  if (content_view)
    [content_view setFrame: frame];
}

- (BOOL)containsXWindow:(Window)w
{
  if (w == XWINDOW)
    return YES;
  return NO;
}

- (void)setXUnmap:(BOOL)flag
{
  // Unmapped means it is no longer visible
  visible = flag;
}

@end
