/* 
   PXKTextFieldCell.m

   NSTextFieldCell for GNUstep GUI X/DPS Backend

   Copyright (C) 1996 Free Software Foundation, Inc.

   Author:  Scott Christley <scottc@net-community.com>
   Date: March 1996
   
   This file is part of the GNUstep GUI X/DPS Backend.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   If you are interested in a warranty or support for this source code,
   contact Scott Christley <scottc@net-community.com> for more information.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 

#include <gnustep/xdps/PXKTextFieldCell.h>
#include <gnustep/xdps/PXKWindow.h>
#include <gnustep/xdps/PXKDPSContext.h>

@implementation PXKTextFieldCell

/* Assumes frame has already been translated to X coordiantes */
- (void)drawBorder:(NSRect)cellFrame
{
  NSDPSContext *cur = [NSDPSContext currentContext];
  Window xWindow;
  Display *xDisplay;
  NSColor *black = [NSColor blackColor];
  NSColor *gray = [NSColor grayColor];
  GC xGC;

  // No border then don't draw it
  if (![self isBordered]) {
    return;
  }

  xWindow = [(PXKWindow *)[control_view window] xWindow];
  xDisplay = [(PXKDPSContext *)cur xDisplay];
  xGC = [(PXKDPSContext *)cur xGC];

  // draw the bottom and right light edges
  [gray set];
  XDrawLine(xDisplay, xWindow, xGC, 
	    cellFrame.origin.x,
	    cellFrame.origin.y + cellFrame.size.height-1,
	    cellFrame.origin.x + cellFrame.size.width-1,
	    cellFrame.origin.y + cellFrame.size.height-1);
  XDrawLine(xDisplay, xWindow, xGC,
	    cellFrame.origin.x + cellFrame.size.width-1,
	    cellFrame.origin.y,
	    cellFrame.origin.x + cellFrame.size.width-1,
	    cellFrame.origin.y + cellFrame.size.height-1);

  // draw the top and left dark edges
  [black set];
  XDrawLine(xDisplay, xWindow, xGC,
	    cellFrame.origin.x,
	    cellFrame.origin.y,
	    cellFrame.origin.x + cellFrame.size.width-2,
	    cellFrame.origin.y);
  XDrawLine(xDisplay, xWindow, xGC,
	    cellFrame.origin.x,
	    cellFrame.origin.y,
	    cellFrame.origin.x,
	    cellFrame.origin.y + cellFrame.size.height-2);
}

/* Assumes frame has already been translated to X coordinates */
- (void)drawText:(NSRect)cellFrame
	 XWindow:(Window)xWindow
	XDisplay:(Display *)xDisplay
{
  PXKDPSContext *cur = [NSDPSContext currentContext];
  float len1;
  float width1;
  int x = 0, y = 0;
  float font_height;
  XFontStruct *win_font;
  GC xgc;
  NSColor *black = [NSColor blackColor];

  (void)load_font(&win_font);
  if (!win_font) {
    NSLog(@"Error loading font.\n");
    return;
  }

  /* need length for both XTextWidth and XDrawString */
  len1 = [contents length];

  /* get string widths for centering */
  width1 = (float)XTextWidth(win_font, [contents cString], len1);
  width1 -= (3 * len1);
  font_height = win_font->ascent + win_font->descent;

  /* If scrollable then we need to show as much of the string as
     possible with the cursor; otherwise, just show as much of the
     string from its beginning.  It only matters if the string
     width is greater than the display field width. */
  if (width1 > (cellFrame.size.width - 6))
    {
      if (cell_scrollable)
	{
	  // I keep subdividing and testing the string length
	  // complexity is O(log(n)) where n is length of string
	  char *cs = [contents cString];
	  BOOL done = NO;
	  unsigned int top_len = (unsigned int)len1;
	  unsigned int bot_len = top_len/2, prev_len = top_len;
	  unsigned int wid;
	  while (!done)
	    {
	      NSDebugLog(@"PXKTextFieldCell: scroll %d %d %f\n", 
		    bot_len, prev_len, len1);
	      if (bot_len >= (prev_len-1))
		  done = YES;
	      else
		{
		  wid = XTextWidth(win_font, cs, bot_len);
		  wid -= (3 * bot_len);
		  // If the width is greater then subdivide again
		  // else increase the size
		  if (wid > (cellFrame.size.width - 6))
		    {
		      unsigned int temp_len = bot_len;
		      bot_len = bot_len - (prev_len - bot_len);
		      prev_len = temp_len;
		    }
		  else
		    {
		      bot_len = bot_len + ((prev_len - bot_len)/2);
		    }
		}
	    }
	  len1 = bot_len;
	  // We may be on the edge so subtract 1 if necessary
	  wid = XTextWidth(win_font, cs, len1);
	  wid -= (3 * len1);
	  if (wid > (cellFrame.size.width - 6))
	    --len1;
	}
      else
	{
	  // I keep subdividing and testing the string length
	  // complexity is O(log(n)) where n is length of string
	  char *cs = [contents cString];
	  BOOL done = NO;
	  unsigned int top_len = (unsigned int)len1;
	  unsigned int bot_len = top_len/2, prev_len = top_len;
	  unsigned int wid;
	  while (!done)
	    {
	      if (bot_len >= (prev_len-1))
		  done = YES;
	      else
		{
		  wid = XTextWidth(win_font, cs, bot_len);
		  wid -= (3 * bot_len);
		  // If the width is greater then subdivide again
		  // else increase the size
		  if (wid > (cellFrame.size.width - 6))
		    {
		      prev_len = bot_len;
		      bot_len = prev_len/2;
		    }
		  else
		    {
		      unsigned int temp_len = bot_len;
		      bot_len = bot_len + ((prev_len - bot_len)/2);
		      prev_len = temp_len;
		    }
		}
	    }
	  NSDebugLog(@"PXKTextFieldCell: no-scroll %d %d\n", 
		     bot_len, prev_len);
	  len1 = bot_len;
	}
      // Recalculate the final width
      width1 = (float)XTextWidth(win_font, [contents cString], len1);
      width1 -= (3 * len1);
    }

  switch(text_align)
    {
    case NSJustifiedTextAlignment:
    case NSNaturalTextAlignment:
    case NSLeftTextAlignment:
      // Upper left corner
      x = (int)cellFrame.origin.x + 3;
      y = (int)(cellFrame.origin.y + (float)win_font->ascent + 3);
      break;

    case NSRightTextAlignment:
      // Upper right corner
      x = (int)cellFrame.origin.x + (cellFrame.size.width - width1) - 3;
      y = (int)(cellFrame.origin.y + (float)win_font->ascent + 3);
      break;

    case NSCenterTextAlignment:
      // Upper center
      x = (int)(cellFrame.origin.x + ((cellFrame.size.width - width1)/2.0));
      y = (int)(cellFrame.origin.y + (float)win_font->ascent + 3);
      break;
    }
  NSDebugLog(@"PXKTextFieldCell text %f %f %d %d %f\n", width1, 
	     font_height, x, y, cellFrame.size.width);

  // Set the color in the current GC
  [text_color set];
  xgc = [(PXKDPSContext *)cur xGC];

  // Draw the text
  XDrawString(xDisplay, xWindow, xgc, x, y, [contents cString], len1);

  // If we are the first responder then draw a cursor
  if ([[control_view window] firstResponder] == control_view)
    {
      [black set];
      XDrawLine(xDisplay, xWindow, xgc, x + width1 + 1, y,
		x + width1 + 1, y - win_font->ascent);
    }
}

// OpenStep drawing methods
- (void)drawInteriorWithFrame:(NSRect)cellFrame
		       inView:(NSView *)controlView
{
  PXKDPSContext *cur = [NSDPSContext currentContext];
  GC xgc;
  Window xWindow = [(PXKWindow *)[control_view window]xWindow];
  Display *xDisplay = [(PXKDPSContext *)cur xDisplay];
  NSRect xframe = [[[controlView window] contentView] bounds];
  NSRect cf = [controlView convertRect: cellFrame toView: nil];

  NSDebugLog(@"PXKTextFieldCell: drawInteriorWithFrame:inView:\n");

  // If the X/Window hasn't been exposed yet
  // then don't do any drawing
  if (![(PXKWindow *)[control_view window] isXExposed])
    return;

  // Flip the coordinate system
  cf.origin.y = xframe.size.height - cf.size.height - cf.origin.y;

  // Set the color in the current GC
  [background_color set];
  xgc = [(PXKDPSContext *)cur xGC];

  // Fill the interior with white
  XFillRectangle(xDisplay, xWindow, xgc,
		 cf.origin.x + 1, cf.origin.y + 1,
		 cf.size.width - 2, cf.size.height - 2);

  // Draw the text
  [self drawText: cf XWindow: xWindow XDisplay: xDisplay];
}

- (void)drawWithFrame:(NSRect)cellFrame
	       inView:(NSView *)controlView
{
  NSRect xframe = [[[controlView window] contentView] bounds];
  NSRect cf = [controlView convertRect: cellFrame toView: nil];

  // If the X/Window hasn't been exposed yet
  // then don't do any drawing
  if (![(PXKWindow *)[controlView window] isXExposed])
    return;

  NSDebugLog(@"PXKTextFieldCell: drawWithFrame:inView:\n");
  // Flip the coordinate system
  cf.origin.y = xframe.size.height - cf.size.height - cf.origin.y;

  // Save last view drawn to
  control_view = controlView;

  // now draw it
  [self drawBorder:cf];
  [self drawInteriorWithFrame: cellFrame inView: controlView];
}

- (void)highlight:(BOOL)lit
	withFrame:(NSRect)cellFrame
	   inView:(NSView *)controlView
{
  [super highlight: lit withFrame: cellFrame inView: controlView];
  [self drawWithFrame: cellFrame inView: controlView];
}

@end
