/* 
   PXKWindow.m

   NSWindow for GNUstep GUI X/DPS Backend.

   Copyright (C) 1996 Free Software Foundation, Inc.

   Author:  Pascal Forget <pascal@wsc.com>
   Date: February 1996
   
   This file is part of the GNUstep GUI X/DPS Backend.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   If you are interested in a warranty or support for this source code,
   contact Scott Christley <scottc@net-community.com> for more information.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 

#include <gnustep/xdps/PXKWindow.h>
#include <gnustep/xdps/PXKView.h>
#include <gnustep/xdps/PXKDPSContext.h>
#include <gnustep/xdps/PXKScreen.h>
#include <gnustep/xdps/PXKColor.h>
#include <gnustep/xdps/PXKApplication.h>
#include <gnustep/xdps/PXKMenu.h>

//
// Backend structure for PXKWindow
//
typedef struct _PXKWindow_struct
{
  Window parent;
  Window xWindow;
  GC context;
  Window menu;
  GC menu_context;
  BOOL is_exposed;
  float menu_height;
  NSSize content_size;
} PXKWindow_struct;

#define XPARENT (((PXKWindow_struct *)be_wind_reserved)->parent)
#define XWINDOW (((PXKWindow_struct *)be_wind_reserved)->xWindow)
#define XGC (((PXKWindow_struct *)be_wind_reserved)->context)
#define XEXPOSE (((PXKWindow_struct *)be_wind_reserved)->is_exposed)
#define XMENU (((PXKWindow_struct *)be_wind_reserved)->menu)
#define XMGC (((PXKWindow_struct *)be_wind_reserved)->menu_context)
#define XMHEIGHT (((PXKWindow_struct *)be_wind_reserved)->menu_height)
#define XCSIZE (((PXKWindow_struct *)be_wind_reserved)->content_size)

@implementation PXKWindow

- (id)initWithContentRect:(NSRect)contentRect
		styleMask:(unsigned int)aStyle
		  backing:(NSBackingStoreType)bufferingType
		    defer:(BOOL)flag
		   screen:aScreen;
{
  // Allocate the back-end structure
  be_wind_reserved = malloc(sizeof(PXKWindow_struct));
  XPARENT = 0;
  XWINDOW = 0;
  XMENU = 0;
  XEXPOSE = NO;
  XMHEIGHT = 0;
  XCSIZE = NSZeroSize;

  [super initWithContentRect:contentRect styleMask:aStyle
	 backing:bufferingType defer:flag screen:aScreen];

  NSDebugLog(@"PXKWindow default initializer\n");

  [self createXWindow: contentRect];

#ifdef TRANSPARENCY
  [content_view setOpaque:YES];
#endif

  NSDebugLog(@"PXKWindow end of default initializer\n");
  return self;
}

- (void)dealloc;
{
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
					xDisplay];

  // Destroy the X Window
  if (XWINDOW)
    {
      XDestroyWindow(xDisplay, XWINDOW);
      XWINDOW = 0;
    }
  // Destroy the X Window
  if (XMENU)
    {
      XDestroyWindow(xDisplay, XMENU);
      XMENU = 0;
    }
  // Destroy the X Window
  if (XPARENT)
    {
      XDestroyWindow(xDisplay, XPARENT);
      XPARENT = 0;
    }

  // Release the back-end structure
  free(be_wind_reserved);

  [super dealloc];
}

- (void)close;
{
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
					xDisplay];

  if (XEXPOSE)
      XUnmapWindow(xDisplay, XPARENT);
  XEXPOSE = NO;
  visible = NO;
}

- (void)setTitle:(NSString *)aString
{
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
					xDisplay];
  XTextProperty windowName;
  char *newTitle;

  [super setTitle: aString];

  if ((window_title) && (XPARENT))
    {
      newTitle = (char *)malloc([window_title length] + 1);
      strcpy(newTitle, [window_title cString]);
      XStringListToTextProperty(&newTitle, 1, &windowName);
      XSetWMName(xDisplay, XPARENT, &windowName);
      XSetWMIconName(xDisplay, XPARENT, &windowName);
      free(newTitle);
    }
}

- (void)orderFront: sender
{
  [self display];
}

- (void)display;
{
  NSDPSContext *cur = [NSDPSContext currentContext];
  Display *xDisplay = [(PXKDPSContext *)cur xDisplay];

  //  XFlush(xDisplay);
  XMapWindow(xDisplay, XPARENT);

  // Don't try to draw until we have been exposed.
  if (XEXPOSE)
    {
      // Draw the menu if we have one
      if (XMHEIGHT)
	{
	  NSApplication *theApp = [NSApplication sharedApplication];
	  NSMenu *m = [theApp mainMenu];

	  if (m)
	    [(PXKMenu *)m drawMenuForWindow: self];
	}

      // Draw the content view
      [content_view display];
    }

  visible = YES;
}

- (void)sendEvent:(NSEvent *)theEvent
{
  NSPoint p;

  switch ([theEvent type])
    {

      //
      // Mouse events
      //
      //
      // Left mouse down
      //
    case NSLeftMouseDown:
      // Find out if the mouse down happenned in the menu
      // or the content view
      p = [theEvent locationInWindow];
      NSLog(@"Got a mouse down at %f %f %f\n", p.x, p.y);
    default:
    }
  [super sendEvent: theEvent];
}

#if 0
- (void)rightMouseDown:(NSEvent *)theEvent;
{
    NSView *hit = [content_view hitTest:[theEvent locationInWindow]];
    
    if (hit == nil) {
    	[self noResponderFor:@selector(rightMouseDown:)];
	return;
    } else {
	if (([self isKeyWindow] == YES) || 
	    ([hit acceptsFirstMouse:theEvent] == YES))
	{
	    [hit rightMouseDown:theEvent];
	}
    }    
}
#endif

- (void)setContentViewSize:(NSSize)aSize
{
  NSRect r;

  r.origin = NSZeroPoint;
  r.size = aSize;
  if (content_view)
    [content_view setFrame: r];
}

- (void)setContentSize:(NSSize)aSize
{
}

- (void)calcContentSize
{
  NSApplication *theApp = [NSApplication sharedApplication];
  NSMenu *m = [theApp mainMenu];
  NSRect mr = NSZeroRect;

  // Determine size of the menu
  if (m)
    mr = [m frame];
  XMHEIGHT = mr.size.height;

  XCSIZE.width = frame.size.width;
  XCSIZE.height = frame.size.height - XMHEIGHT;
}

- (void)setFrame:(NSRect)frameRect display:(BOOL)flag;
{
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext] 
					xDisplay];
  NSRect xframe = NSZeroRect;
  float x, y;

  frame = frameRect;

  // Adjust the frame to device coordinates
  PSdtransform(frame.size.width, frame.size.height, &x, &y);
  if (x < 0) x = -x;
  if (y < 0) y = -y;
  NSDebugLog(@"transform %f %f %f %f\n", frame.size.width, 
	     frame.size.height, x, y);
  xframe.size.width = x;
  xframe.size.height = y;

  [self calcContentSize];

  // Resize the parent first
  XMoveResizeWindow(xDisplay, XPARENT,
		    0, 0, xframe.size.width, xframe.size.height);
  [self setContentXFrames];

  [self setContentViewSize: XCSIZE];
    
  if (flag)
    [self display];
}

//
// GNUstep backend methods
//
- (void)captureMouse:sender
{
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
					xDisplay];
  int ret = XGrabPointer(xDisplay, [self xWindow], False,
			 PointerMotionMask | ButtonReleaseMask,
			 GrabModeAsync, GrabModeAsync, 
			 None, None, CurrentTime);

  if (ret != GrabSuccess)
    NSDebugLog(@"Failed to grab pointer\n");
  else
    {
      NSApplication *theApp = [NSApplication sharedApplication];
      NSDebugLog(@"Grabbed pointer\n");
      [(PXKApplication *)theApp setGrabXWindow: [self xWindow]];
    }
}

- (void)releaseMouse:sender
{
  NSApplication *theApp = [NSApplication sharedApplication];
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
					xDisplay];
  XUngrabPointer(xDisplay, CurrentTime);
  [(PXKApplication *)theApp setGrabXWindow: 0];
}

@end

static char icon_bitmap_bits2[] = {
   0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

//
// Methods for XWindows implementation
//
@implementation PXKWindow (GNUstepXDPS)

- (void)createXWindow:(NSRect) rect
{
  XColor c1;
  NSColor *c;
  XWMHints *wm_hints;
  XGCValues values;
  unsigned long valuemask = (GCForeground | GCBackground);
  int xScreen = [(PXKScreen *)[NSScreen mainScreen] xScreen];
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
					xDisplay];
  NSApplication *theApp = [NSApplication sharedApplication];
  NSMenu *m = [theApp mainMenu];
  NSRect mr = NSZeroRect, pr = rect;

  if (!(wm_hints = XAllocWMHints())) {
    NSLog(@"PXKWindow: Failure allocating memory with XAllocWMHints()\n");
    exit(0);
  }

  // Window background color
  c = [self backgroundColor];
  if (([[c colorSpaceName] compare: NSCalibratedRGBColorSpace] 
       == NSOrderedSame) ||
      ([[c colorSpaceName] compare: NSDeviceRGBColorSpace]
       == NSOrderedSame))
    {
      // Create an XColor from the NSColor
      c1.red = (unsigned short)(65535 * [c redComponent]);
      c1.green = (unsigned short)(65535 * [c greenComponent]);
      c1.blue = (unsigned short)(65535 * [c blueComponent]);
    }
  else if (([[c colorSpaceName] compare: NSCalibratedWhiteColorSpace] 
	    == NSOrderedSame) ||
	   ([[c colorSpaceName] compare: NSDeviceWhiteColorSpace]
	    == NSOrderedSame))
    {
      // Create an XColor from the NSColor
      c1.red = (unsigned short)(65535 * [c whiteComponent]);
      c1.green = (unsigned short)(65535 * [c whiteComponent]);
      c1.blue = (unsigned short)(65535 * [c whiteComponent]);
    }
  else
    {
      // If not a colorspace we can easily convert then
      // then just make the background gray.
      c1.red = 32000;
      c1.green = 32000;
      c1.blue = 32000;
    }
  XAllocColor(xDisplay, XDefaultColormap(xDisplay, xScreen), &c1);

  // Figure out the size of the parent window by adding the height
  // of the content rect with the menu height
  if (m)
    {
      mr = [m frame];
      pr.size.height += mr.size.height;
    }

  /* Create the parent X window */
  XPARENT = XCreateSimpleWindow(xDisplay, RootWindow(xDisplay, xScreen),
				0, 0,
				1000, 1000,
				0, BlackPixel(xDisplay, xScreen),
				c1.pixel);

  /* Create the menu X window */
  XMENU = XCreateSimpleWindow(xDisplay, XPARENT, 0, 0,
			      1000, 200,
			      0, BlackPixel(xDisplay, xScreen),
			      c1.pixel);
  /* Set the event mask */
  XSelectInput(xDisplay, XMENU,
	       ExposureMask | KeyPressMask | KeyReleaseMask |
	       ButtonPressMask | ButtonReleaseMask | 
	       ButtonMotionMask | StructureNotifyMask |
	       PointerMotionMask | EnterWindowMask |
	       LeaveWindowMask | StructureNotifyMask | FocusChangeMask |
	       PropertyChangeMask | ColormapChangeMask | KeymapStateMask |
	       VisibilityChangeMask);

  // Create a GC for the menu
  values.foreground = BlackPixel(xDisplay, xScreen);
  values.background = WhitePixel(xDisplay, xScreen);
  XMGC = XCreateGC(xDisplay, XMENU, valuemask, &values);

  XWINDOW = XCreateSimpleWindow(xDisplay, XPARENT,
				0,
				200,
				1000,
				800,
				0, BlackPixel(xDisplay, xScreen),
				c1.pixel);

  // Create a GC for the content view
  values.foreground = BlackPixel(xDisplay, xScreen);
  values.background = WhitePixel(xDisplay, xScreen);
  XGC = XCreateGC(xDisplay, XWINDOW, valuemask, &values);

  /* Set the event mask */
  XSelectInput(xDisplay, XWINDOW,
	       ExposureMask | KeyPressMask | KeyReleaseMask |
	       ButtonPressMask | ButtonReleaseMask | 
	       ButtonMotionMask | StructureNotifyMask |
	       PointerMotionMask | EnterWindowMask |
	       LeaveWindowMask | StructureNotifyMask | FocusChangeMask |
	       PropertyChangeMask | ColormapChangeMask | KeymapStateMask |
	       VisibilityChangeMask);

  /* Give the window manager some hints, including our icon */
  wm_hints->initial_state = NormalState;
  wm_hints->input = True;
  wm_hints->icon_pixmap = XCreateBitmapFromData(xDisplay, XPARENT,
						icon_bitmap_bits2, 
						48,48);
    
  wm_hints->flags = StateHint | IconPixmapHint | InputHint;
  XSetWMHints(xDisplay, XPARENT, wm_hints);

  /* Set the event mask */
  XSelectInput(xDisplay, XPARENT,
	       ExposureMask | KeyPressMask | KeyReleaseMask |
	       ButtonPressMask | ButtonReleaseMask | 
	       ButtonMotionMask | StructureNotifyMask |
	       PointerMotionMask | EnterWindowMask |
	       LeaveWindowMask | StructureNotifyMask | FocusChangeMask |
	       PropertyChangeMask | ColormapChangeMask | KeymapStateMask |
	       VisibilityChangeMask);

  XMapSubwindows(xDisplay, XPARENT);
}

- (Window)xWindow
{
  return XWINDOW;
}

- (GC)xGC
{
  return XGC;
}

- (Window)xParentWindow;
{
  return XPARENT;
}

- (Window)xMenuWindow;
{
  return XMENU;
}

- (GC)xMenuGC
{
  return XMGC;
}

- (void)setXExposed:(BOOL)flag
{
  XEXPOSE = flag;
}

- (BOOL)isXExposed
{
  return XEXPOSE;
}

- (void)setContentXFrames
{
  Display *xDisplay = [(PXKDPSContext *)[NSDPSContext currentContext]
					xDisplay];
  NSRect mframe = NSZeroRect, cframe = NSZeroRect;
  float x, y;

  // Adjust the menu frame to device coordinates
  PSdtransform(frame.size.width, XMHEIGHT, &x, &y);
  if (x < 0) x = -x;
  if (y < 0) y = -y;
  NSDebugLog(@"menu transform %f %f %f %f\n", frame.size.width, 
	     XMHEIGHT, x, y);
  mframe.size.width = x;
  mframe.size.height = y;

  // Resize the menu
  if ((XMENU) && (XMHEIGHT != 0))
    {
      XMoveResizeWindow(xDisplay, XMENU,
			0, 0, mframe.size.width, mframe.size.height);
      NSDebugLog(@"resize XMENU\n");
    }

  // Adjust the content frame to device coordinates
  PSdtransform(XCSIZE.width, XCSIZE.height, &x, &y);
  if (x < 0) x = -x;
  if (y < 0) y = -y;
  NSDebugLog(@"content transform %f %f %f %f\n", XCSIZE.width, 
	     XCSIZE.height, x, y);
  cframe.origin.y = mframe.size.height;
  cframe.size.width = x;
  cframe.size.height = y;

  // Resize the content view
  if (XWINDOW)
    {
      XMoveResizeWindow(xDisplay, XWINDOW,
			0, cframe.origin.y, 
			cframe.size.width, cframe.size.height);
      NSDebugLog(@"resize XWINDOW\n");
    }
}

- (void)setFrameFromXFrame:(NSRect)rect
{
  float x, y;

  // Adjust the frame to user coordinates
  PSidtransform(rect.size.width, rect.size.height, &x, &y);
  if (x < 0) x = -x;
  if (y < 0) y = -y;
  NSDebugLog(@"transform %f %f %f %f\n", rect.size.width, 
	     rect.size.width, x, y);
  frame.size.width = x;
  frame.size.height = y;

  [self calcContentSize];
  [self setContentXFrames];

  // Inform our content view
  [self setContentViewSize: XCSIZE];
}

- (BOOL)containsXWindow:(Window)w
{
  if (w == XWINDOW)
    return YES;
  if (w == XPARENT)
    return YES;
  //  if (w == XMENU)
  //    return YES;
  return NO;
}

@end

